@@ -2,7 +2,7 @@
 #
 #   This file is part of PyBuilder
 #
-#   Copyright 2011-2014 PyBuilder Team
+#   Copyright 2011-2020 PyBuilder Team
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
@@ -18,20 +18,26 @@
 
 import os
 import shutil
+from functools import partial
 
-from pybuilder.core import init, task, description, depends
+from pybuilder.core import init, task, description, depends, optional
+from pybuilder.utils import jp
 
 
 @init
 def init(project):
     project.set_property("dir_target", "target")
-    project.set_property("dir_reports", "$dir_target/reports")
-    project.set_property("dir_logs", "$dir_target/logs")
+    project.set_property("dir_reports", jp("$dir_target", "reports"))
+    project.set_property("dir_logs", jp("$dir_target", "logs"))
 
-    def write_report(file, *content):
-        with open(project.expand_path("$dir_reports", file), "w") as report_file:
-            report_file.writelines(content)
-    project.write_report = write_report
+    project.set_property_if_unset("remote_debug", 0)
+
+    project.write_report = partial(write_report, project)
+
+
+def write_report(project, file, *content):
+    with open(project.expand_path("$dir_reports", file), "w") as report_file:
+        report_file.writelines(content)
 
 
 @task
@@ -44,7 +50,7 @@ def clean(project, logger):
 
 @task
 @description("Prepares the project for building.")
-def prepare(project, logger):
+def prepare(project, logger, reactor):
     target_directory = project.expand_path("$dir_target")
     if not os.path.exists(target_directory):
         logger.debug("Creating target directory %s", target_directory)
@@ -55,6 +61,9 @@ def prepare(project, logger):
         logger.debug("Creating reports directory %s", reports_directory)
         os.mkdir(reports_directory)
 
+    reactor.python_env_registry["pybuilder"].install_dependencies(project.plugin_dependencies,
+                                                                  package_type="plugin")
+
 
 @task
 @depends(prepare)
@@ -71,7 +80,7 @@ def run_unit_tests():
 
 
 @task
-@depends(run_unit_tests)
+@depends(compile_sources, optional(run_unit_tests))
 @description("Packages the application.")
 def package():
     pass
@@ -85,14 +94,31 @@ def run_integration_tests():
 
 
 @task
-@depends(run_integration_tests)
+@depends(optional(run_integration_tests))
 @description("Verifies the project and possibly integration tests.")
 def verify():
     pass
 
 
 @task
-@depends(verify)
+@depends(package, optional(verify))
 @description("Publishes the project.")
 def publish():
     pass
+
+
+@task
+@depends(package, optional(publish))
+@description("Installs the published project.")
+def install():
+    pass
+
+
+@task(description="Print the module path.")
+def print_module_path(project):
+    print(project.expand_path("$dir_source_main_python"))
+
+
+@task(description="Print the script path.")
+def print_scripts_path(project):
+    print(project.expand_path("$dir_source_main_scripts"))
