@@ -2,7 +2,7 @@
 #
 #   This file is part of PyBuilder
 #
-#   Copyright 2011-2014 PyBuilder Team
+#   Copyright 2011-2020 PyBuilder Team
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
@@ -24,33 +24,33 @@
     https://github.com/timothycrosley/frosted
 """
 
-__author__ = 'Maximilien Riehl'
-
 from pybuilder.core import after, task, init, use_plugin, depends
 from pybuilder.errors import BuildFailedException
-from pybuilder.utils import assert_can_execute
 from pybuilder.pluginhelper.external_command import ExternalCommandBuilder
+from pybuilder.utils import tail_log
 
+__author__ = "Maximilien Riehl"
 
 use_plugin("python.core")
 
 
 @init
 def initialize_frosted_plugin(project):
-    project.build_depends_on("frosted")
+    project.plugin_depends_on("frosted")
+
     project.set_property("frosted_break_build", False)
     project.set_property("frosted_include_test_sources", False)
     project.set_property("frosted_include_scripts", False)
 
 
 @after("prepare")
-def assert_frosted_is_executable(logger):
+def assert_frosted_is_executable(project, logger, reactor):
     """ Asserts that the frosted script is executable. """
     logger.debug("Checking if frosted is executable.")
 
-    assert_can_execute(command_and_arguments=["frosted", "--version"],
-                       prerequisite="frosted (PyPI)",
-                       caller="plugin python.frosted")
+    reactor.python_env_registry["pybuilder"].verify_can_execute(command_and_arguments=["frosted", "--version"],
+                                                                prerequisite="frosted (PyPI)",
+                                                                caller="plugin python.frosted")
 
 
 @task
@@ -62,9 +62,9 @@ def analyze(project, logger):
     verbose = project.get_property("verbose")
     project.set_property_if_unset("frosted_verbose_output", verbose)
 
-    command = ExternalCommandBuilder('frosted', project)
-    for ignored_error_code in project.get_property('frosted_ignore', []):
-        command.use_argument('--ignore={0}'.format(ignored_error_code))
+    command = ExternalCommandBuilder("frosted", project)
+    for ignored_error_code in project.get_property("frosted_ignore", []):
+        command.use_argument("--ignore={0}".format(ignored_error_code))
 
     include_test_sources = project.get_property("frosted_include_test_sources")
     include_scripts = project.get_property("frosted_include_scripts")
@@ -77,7 +77,8 @@ def analyze(project, logger):
     count_of_errors = len(result.error_report_lines)
 
     if count_of_errors > 0:
-        logger.error('Errors while running frosted, see {0}'.format(result.error_report_file))
+        logger.error("Errors while running frosted. See %s for full details:\n%s" % (
+            result.error_report_file, tail_log(result.error_report_file)))
 
     if count_of_warnings > 0:
         if project.get_property("frosted_break_build"):
